/* global win, error_handler, draw_full_frame, browser, gen_nodes */
"use strict";

function $(id) {
	return document.getElementById(id);
}

var crc32 = (function() {
	var crc_table = [], n = 0, c, k;

	for (; n < 256; ++n) {
		for (c = n, k = 0; k < 8; ++k) {
			c = (c & 1) ? 0xedb88320 ^ (c >>> 1) : c >>> 1;
		}

		crc_table[n] = c;
	}

	return function(s) {
		var crc = -1, n = 0, l = s.length;

		while (n < l) {
			crc = crc_table[(crc ^ s.charCodeAt(n++)) & 0xff] ^ (crc >>> 8);
		}

		return crc ^ -1;
	};
})();

function BinaryTools(data) {
	this.pos = 0;
	this.data = data;
	this.length = data.length;
	this.littleEndian = false;
	this.zeropad = "00000000";

	this.readByte = function(pos) {
		return this.data.charCodeAt(
			pos === void 0 || pos < 0 ? this.pos++ : pos
		) & 0xff;
	};

	this.readString = function(length, pos) {
		if (pos === void 0 || pos < 0) {
			pos = this.pos;
			this.pos += length || 0;
		}

		return this.data.substr(pos, length || 0);
	};

	this.readBits = function(length, pos) {
		var curbyte, bitarray = [], i = 0;

		while (i++ < length) {
			curbyte = this.readByte(pos).toString(2);
			bitarray.push(
				this.zeropad.substr(0, this.zeropad.length - curbyte.length) + curbyte
			);
		}

		return (this.littleEndian ? bitarray.reverse() : bitarray).join("");
	};

	this.readInt = function(bytes, pos) {
		if (bytes < 1) {
			return null;
		}

		if (pos === void 0 || pos < 0) {
			pos = this.pos;
			this.pos += bytes;
		}

		var i, integer = 0;

		if (this.littleEndian) {
			i = bytes - 1;

			while (i >= 0) {
				integer += this.readByte(pos + i) << (i-- * 8);
			}
		}
		else {
			i = 0;

			while (i < bytes) {
				integer += this.readByte(pos + i++) << ((bytes - i) * 8);
			}
		}

		return integer >>> 0;
	};

	this.intToBytes = function(integer, i) {
		if (!i) {
			i = 4;
		}

		var bytes = [];

		while (i--) {
			bytes.push((integer >> (i * 8)) & 0xff);
		}

		return String.fromCharCode.apply(
			null,
			this.littleEndian ? bytes.reverse() : bytes
		);
	};
}

if (!win.opera) {
	var b64chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	win.btoa = function(b64str) {
		var c1, c2,
			pos = 0,
			res = "",
			mod = b64str.length % 3,
			l = b64str.length - mod;

		while (pos < l) {
			c1 = b64str.charCodeAt(pos++) & 0xff;
			c2 = b64str.charCodeAt(pos++) & 0xff;
			res += b64chars[c1 >> 2];
			res += b64chars[((c1 & 3) << 4) | (c2 >> 4)];
			c1 = b64str.charCodeAt(pos++) & 0xff;
			res += b64chars[((c2 & 0x0f) << 2) | (c1 >> 6)];
			res += b64chars[c1 & 0x3f];
		}

		if (mod) {
			c1 = b64str.charCodeAt(pos++) & 0xff;
			res += b64chars[c1 >> 2];

			if (mod === 1) {
				res += b64chars[(c1 & 3) << 4];
				res += "==";
			}
			else {
				c2 = b64str.charCodeAt(pos++) & 0xff;
				res += b64chars[((c1 & 3) << 4) | (c2 >> 4)];
				res += b64chars[(c2 & 0x0f) << 2];
				res += "=";
			}
		}

		return res;
	};
}

var maxSize = 20 * 1024 * 1024;
var xhr = new win.XMLHttpRequest;

xhr.overrideMimeType("text/plain; charset=x-user-defined");
xhr.onreadystatechange = function() {
	if (this.readyState === 2) {
		var length = this.getResponseHeader("Content-Length");

		if (length && length > maxSize) {
			this.abort();
			error_handler("Image is too large...");
			return;
		}
	}

	if (this.responseText) {
		if (!this.img_type) {
			// PNG or GIF or WEBP signature
			img_head = /^(?:\x89(PNG)\r\n\x1a\n|(GIF)8[79]a|RIFF....(WEBP)VP8X)/;

			this.img_type = this.responseText.match(img_head);

			// seems like in some cases a character encoding is applied anyway,
			// however it's enough to check the signature only
			if (!this.img_type && (this.responseText[1] === "P" && (chunkSize = 8)
					|| this.responseText[0] === "G" && (chunkSize = 6)
					|| this.responseText[0] === "R" && (chunkSize = 16))) {
				this.img_type = this.responseText.slice(0, chunkSize).split("");

				for (i = 0; i < this.img_type.length; ++i) {
					this.img_type[i] = this.img_type[i].charCodeAt(0) & 0xff;
				}

				this.img_type = String.fromCharCode.apply(null, this.img_type).match(img_head);
			}

			if (this.img_type) {
				this.img_type = this.img_type[2] || this.img_type[1] || this.img_type[3];
			}
			else {
				this.abort();
				this.onprogress = null;
				error_handler("Not animated...");
				return;
			}
		}

		if (this.readyState !== 4/* || this.status !== 200*/) {
			return;
		}

		if (this.responseText.length > maxSize) {
			this.onprogress = null;
			error_handler("Image is too large...");
			return;
		}

		var
			i, IHDR, chunkType, chunkSize,
			frames = [],
			bin = new BinaryTools(this.responseText),
			animation = {},
			img_head = "";

		if (this.img_type === "PNG") {
			// https://wiki.mozilla.org/APNG_Specification

			// skip signature
			bin.pos = 8;

			while (bin.pos < bin.length) {
				// read the chunk type
				chunkType = bin.readString(4, bin.pos + 4);

				// console.log(i, chunkType, bin.readInt(4, bin.pos));

				if (chunkType === "IHDR") {
					// skip chunk length and name
					bin.pos += 8;
					animation.width = bin.readInt(4);
					animation.height = bin.readInt(4);
					IHDR = bin.readString(5);
					// skip crc
					bin.pos += 4;
				}
				else if (!frames.length && chunkType === "acTL") {
					animation.num_frames = bin.readInt(4, bin.pos + 8);
					animation.num_plays = bin.readInt(4, bin.pos + 12);
					bin.pos += 20;
				}
				else if (chunkType === "fcTL") {
					bin.pos += 12;
					i = frames.length;

					frames.push({
						width: bin.readInt(4),
						height: bin.readInt(4),
						x_offset: bin.readInt(4),
						y_offset: bin.readInt(4),
						delay: 1000 * bin.readInt(2) / (bin.readInt(2) || 100),
						dispose_op: bin.readInt(1),
						blend_op: bin.readInt(1),
						data: "IDAT"
					});

					if (i === 0 && frames[i].dispose_op === 2) {
						frames[i].dispose_op = 1;
					}

					// skip crc
					bin.pos += 4;
					// console.log(frames[frames.length - 1]);
				}
				else if (chunkType === "IDAT" || chunkType === "fdAT") {
					if (!animation.num_frames || animation.num_frames < 2) {
						break;
					}
					// IDAT without acTL chunk should be ignored
					else if (chunkType === "IDAT" && !frames.length) {
						console.log("PNG: ignoring IDAT from animation...");
						bin.pos += bin.readInt(4, bin.pos) + 12;
						continue;
					}

					i = (chunkType === "fdAT" ? 4 : 0);
					// read the length, also skip the sequence number from fdAT
					chunkSize = bin.readInt(4) - i;
					// skip chunk type, also the sequence number
					// in case of the fdAT chunk type
					bin.pos += 4 + i;

					frames[frames.length - 1].data += bin.readString(chunkSize);

					// skip crc
					bin.pos += 4;
				}
				else if (!frames.length) {
					// read the full chunk (length + chunk type + chunkdate + crc)
					img_head += bin.readString(bin.readInt(4, bin.pos) + 12);
				}
				else {
					bin.pos += bin.readInt(4, bin.pos) + 12;
				}
			}
		}
		else if (this.img_type === "GIF") {
			// http://www.w3.org/Graphics/GIF/spec-gif89a.txt

			bin.littleEndian = true;
			bin.skipSubBlock = function() {
				do {
					this.pos += this.readInt(1, this.pos) + 1;

					if (bin.pos >= bin.length) {
						throw this.img_type + ": end reached...";
					}
				} while (this.readInt(1, this.pos) !== 0x00);
				++this.pos;
			};

			// skip signature
			bin.pos = 6;

			// Logical Screen Descriptor
			animation.width = bin.readInt(2);
			animation.height = bin.readInt(2);

			bin.packed = bin.readBits(1, 10);
			bin.packed = (bin.packed[0] === "1" ?
				3 * 1 << (parseInt(bin.packed.slice(-3), 2) + 1) : 0);

			img_head += bin.readString(3 + bin.packed);
			i = 0; // frame index

			while (bin.pos < bin.length) {
				chunkType = bin.readString(1);
				// console.log("sentinel: " + chunkType);

				// Extension block (0x21, an exclamation point '!')
				if (chunkType === "!") { // optional
					chunkType = bin.readInt(1);
					// console.log("ext_label: " + chunkType.toString(16));

					// Plain Text Extension
					if (chunkType === 0x01) {
						// no support for this, so skip
						bin.pos += 13;
						bin.skipSubBlock();
					}
					// Graphics Control Extension
					else if (chunkType === 0xf9) {
						bin.packed = bin.readBits(1, bin.pos + 1);

						frames.push({
							delay: (10 * bin.readInt(2, bin.pos + 2)) || 100,
							// -1 in order to match PNG's indexes, also treat 0 as 1
							dispose_op: Math.max(0, parseInt(bin.packed.slice(3, -2), 2) - 1),
							data: bin.readString(4, bin.pos - 2) + // sentinel + ext_label
										bin.intToBytes(0, 2) + // with zero delay
										// transparent color index + block terminator
										bin.readString(2, bin.pos + 4)
						});

						bin.pos += 6;
					}
					// Comment Block Extension
					else if (chunkType === 0xfe) {
						bin.skipSubBlock();
					}
					// Application Block Extension
					else if (chunkType === 0xff) {
						bin.pos += 12;
						bin.skipSubBlock();
					}
					// try to skip anything else
					else {
						bin.pos = bin.data.indexOf("\x00", bin.pos);

						if (bin.pos === -1) {
							bin.pos = bin.length;
						}
						else {
							++bin.pos;
						}
					}
				}
				// Image (0x2C, a comma ',')
				else if (chunkType === ",") {
					if (!frames[i]) {
						frames.push({
							delay: 100,
							dispose_op: 0,
							// sentinel, ext_label, block size, packed field, delay,
							// transparent color index, terminator
							data: "\x21\xf9\x04\x00\x00\x00\x00\x00"
						});
					}

					frames[i].x_offset = bin.readInt(2);
					frames[i].y_offset = bin.readInt(2);

					bin._pos = bin.pos;

					frames[i].width = bin.readInt(2);
					frames[i].height = bin.readInt(2);

					bin.packed = bin.readBits(1);
					bin.pos += (bin.packed[0] === "1"
							? 3 * (1 << (parseInt(bin.packed.slice(-3), 2) + 1))
							: 0);

					++bin.pos; // LZW minimum code size

					bin.skipSubBlock();

					frames[i]._data = [bin.pos - bin._pos, bin._pos];

					++i;
				}
				// Trailer (0x3B, a semi-colon ';') or something else
				else {
					break;
				}
			}
		}
		else if (this.img_type === "WEBP") {
			// https://developers.google.com/speed/webp/docs/riff_container

			bin.littleEndian = true;

			bin.pos = 20; // WebP file header + VP8X and its chunk size
			bin.packed = bin.readBits(1); // VP8X flags

			if (bin.packed[6] === "0") { // animation flag
				this.onprogress = null;
				error_handler(this.img_type + ": not animated...");
				return;
			}

			bin.pos += 3; // reserved 24 bits
			animation.width = bin.readInt(3) + 1; // canvas width
			animation.height = bin.readInt(3) + 1; // canvas height

			while (bin.pos < bin.length) {
				chunkType = bin.readString(4);

				if (chunkType === "ANIM") {
					// skip Background Color: 32 bits + Loop Count: 16 bits
					bin.pos += 4 + bin.readInt(4); // chunk size + chunk payload
					animation.ANIM = true;
				}
				else if (chunkType === "ANMF") {
					if (!animation.ANIM) {
						this.onprogress = null;
						error_handler(this.img_type + ": ANIM chunk not found!");
						return;
					}

					i = frames.length;
					// size of the Frame Data
					chunkSize = bin.readInt(4);
					// if the chunk size is odd, then a padding byte is added
					chunkSize += chunkSize & 1;
					// remove the size of the ANMF properties listed below
					chunkSize -= 16;

					frames.push({
						x_offset: bin.readInt(3) * 2,
						y_offset: bin.readInt(3) * 2,
						width: bin.readInt(3) + 1,
						height: bin.readInt(3) + 1,
						delay: bin.readInt(3),
						dispose_op: bin.readInt(1) & 1,
						data: ""
					});

					bin._pos = bin.pos + chunkSize;

					while (bin.pos < bin._pos) {
						chunkType = bin.readString(4);
						chunkSize = bin.readInt(4);
						chunkSize += chunkSize & 1;

						// should exist before the 'VP8 ' chunk only
						/*if (chunkType === "ALPH") {
							if (bin.readString(4, bin.pos + chunkSize) === "VP8 ") {
								// the whole ALPH chunk
								frames[i].data += bin.readString(8, bin.pos - 8) + bin.readString(chunkSize);
							}
							else {
								bin.pos += chunkSize;
							}
						}
						else */if (chunkType === "VP8 " || chunkType === "VP8L") {
							// chunk type + chunk size + VP8? bitstream;
							frames[i].data += bin.readString(8, bin.pos - 8) + bin.readString(chunkSize);
						}
						else {
							bin.pos += chunkSize;
						}
					}

					// frames[i].data += bin.readString(chunkSize);
				}
				else { // skip chunk
					bin.pos += 4 + bin.readInt(4);
				}
			}
		}

		if (frames.length > 1) {
			var
				generateImageSRC,
				img = new Image,
				canvas = document.createElement("canvas"),
				ctx = canvas.getContext("2d"),
				wrap, speed, current_frame;

			canvas.width = animation.width;
			canvas.height = animation.height;

			if (this.img_type === "PNG") {
				chunkSize = bin.intToBytes(13);
				var img_end = bin.readString(12, bin.length - 12);

				generateImageSRC = function() {
					var frm = frames[frames.idx],
						IHDR_ = "IHDR" +
										bin.intToBytes(frm.width) + bin.intToBytes(frm.height) +
										IHDR;

					IHDR_ += bin.intToBytes(crc32(IHDR_));

					return "data:image/png;base64," + win.btoa(
						"\x89PNG\r\n\x1a\n" +
						chunkSize + IHDR_ +
						img_head +
						bin.intToBytes(frm.data.length - 4) + frm.data +
						bin.intToBytes(crc32(frm.data)) +
						img_end
					);
				};
			}
			else if (this.img_type === "GIF") {
				chunkSize = bin.intToBytes(0, 2);
				chunkSize = "," + chunkSize + chunkSize;

				generateImageSRC = function() {
					var frm = frames[frames.idx];

					return "data:image/gif;base64," + win.btoa(
						"GIF89a" +
						bin.intToBytes(frm.width, 2) + bin.intToBytes(frm.height, 2) +
						img_head +
						frm.data + chunkSize +
						bin.readString(frm._data[0], frm._data[1]) +
						";"
					);
				};
			}
			else if (this.img_type === "WEBP") {
				generateImageSRC = function() {
					var frm = frames[frames.idx];

					return "data:image/webp;base64," + win.btoa(
						"RIFF" + bin.intToBytes(frm.data.length + 4) + "WEBP" + frm.data
					);
				};
			}

			img.onerror = function() {
				wrap.textContent = xhr.img_type + ": frame (" + (frames.idx + 1) + ") couldn't be parsed!";
				wrap.appendChild(this);
			};

			img.onload = function() {
				if (draw_full_frame) {
					var frame = frames[frames.idx];
					var prev = frames[frames.idx - 1];

					if (prev) {
						if (prev.dispose_op === 1) {
							ctx.clearRect(
								prev.x_offset, prev.y_offset,
								prev.width, prev.height
							);
						}
						else if (prev.dispose_op === 2) {
							ctx.putImageData(
								this.prev_dispose_data,
								prev.x_offset, prev.y_offset
							);
						}
					}

					if (frame.dispose_op === 2) {
						this.prev_dispose_data = ctx.getImageData(
							frame.x_offset, frame.y_offset,
							frame.width, frame.height
						);
					}

					if (frame.blend_op === 0) { // only for PNG
						ctx.clearRect(
							frame.x_offset, frame.y_offset,
							frame.width, frame.height
						);
					}

					ctx.drawImage(
						this,
						frame.x_offset, frame.y_offset
					);
				}

				var c = canvas.cloneNode(false);

				if (draw_full_frame) {
					c.getContext("2d").putImageData(
						ctx.getImageData(0, 0, canvas.width, canvas.height),
						0, 0
					);
				}
				else {
					c.width = this.width;
					c.height = this.height;
					c.getContext("2d").drawImage(this, 0, 0);
				}

				wrap.appendChild(c);

				if (c.previousElementSibling) {
					c.previousElementSibling.style.display = "";
					++current_frame.value;
				}

				c.style.display = "inline-block";
				this.removeAttribute("src");

				++frames.idx;
				processNextFrame();
			};

			var processNextFrame = function() {
				if (frames.idx < frames.length) {
					img.src = generateImageSRC();
					frames[frames.idx].data = null;
				}
				else {
					img.onload = null;
					img = null;
					done();
					error_handler(null);
				}
			};

			var done = function() {
				wrap.current = 0;
				wrap.children[wrap.children.length - 1].style.display = "";
				wrap.children[wrap.current].style.display = "inline-block";
				current_frame.value = 1;
				current_frame.oninput = function() {
					if (parseFloat(speed.value, 10) !== 0) {
						wrap.stop();
					}

					if (wrap.classList.contains("showall")) {
						wrap.onmouseup({button: 0});
					}

					wrap.step(null);
				};

				speed.oninput = function() {
					if (parseFloat(speed.value, 10) === 0) {
						wrap.stop();
						wrap.addEventListener(browser["wheel"], wrap.wheeler, false);
					}
					else {
						wrap.classList.remove("showall");
						wrap.removeEventListener(browser["wheel"], wrap.wheeler, false);
						wrap.animate();
					}
				};

				wrap.wheeler = function(e) {
					wrap.step((e.deltaY || -e.wheelDelta) > 0);
					e.preventDefault();
				};

				wrap.addEventListener(browser["wheel"], wrap.wheeler, false);
				current_frame.addEventListener(browser["wheel"], wrap.wheeler, false);

				wrap.onmouseup = function(e) {
					if (e.button !== 0) {
						return;
					}

					if (e.ctrlKey) {
						if (e.target.toDataURL && speed.value < 1) {
							var img = new Image;
							img.src = e.target.toDataURL();
							this.replaceChild(img, e.target);
							wrap.step(null);
						}

						return;
					}

					this.stop();

					if (this.classList.contains("showall")) {
						if (e.target && e.target.parentNode === wrap) {
							current_frame.value = [].indexOf.call(wrap.childNodes, e.target) + 1;
							wrap.step(null);
						}

						this.addEventListener(browser["wheel"], this.wheeler, false);
					}
					else {
						this.removeEventListener(browser["wheel"], this.wheeler, false);
					}

					this.classList.toggle("showall");
				};

				wrap.animate = function() {
					clearTimeout(wrap.timeout);

					wrap.step(parseFloat(speed.value, 10) < 0);

					if (parseFloat(speed.value, 10)) {
						wrap.timeout = setTimeout(
							wrap.animate,
							frames[wrap.current].delay / Math.abs(parseFloat(speed.value, 10) || 0)
						);
					}
				};

				wrap.stop = function() {
					clearTimeout(this.timeout);
					speed.value = 0;
				};

				wrap.step = function(backward) {
					wrap.children[wrap.current].style.display = "";

					if (backward === null) {
						wrap.current = Math.max(
							1,
							Math.min(parseInt(current_frame.value, 10) || 1, current_frame.max)
						) - 1;
					}
					else {
						if (backward) {
							--wrap.current;
						}
						else {
							++wrap.current;
						}
					}

					if (wrap.current >= wrap.children.length) {
						wrap.current = 0;
					}
					else if (wrap.current < 0) {
						wrap.current = wrap.children.length - 1;
					}

					wrap.children[wrap.current].style.display = "inline-block";
					current_frame.value = wrap.current + 1;
					current_frame.nextElementSibling.value = current_frame.value + " / " + frames.length;
				};
			};

			wrap = document.body;
			wrap.className = "frames";
			wrap.textContent = "";
			gen_nodes(wrap, [
				{tag: "a", attrs: {"class": "back", href: win.location.href}, text: "\u2190"}, " ",
				{tag: "input", attrs: {type: "number", id: "speed", size: 6, value: 0, step: 0.5, min: -10, max: 10}}, "x ",
				{tag: "input", attrs: {type: "range", id: "current_frame", style: "width: 500px; vertical-align: middle;", size: 6, value: 1, step: 1, min: 1}}, " ",
				{tag: "output", attrs: {"for": "current_frame"}, text: "1 / " + frames.length}, " ",
				{tag: "div", attrs: {id: "frames"}}
			]);

			wrap = $("frames");
			speed = $("speed");
			current_frame = $("current_frame");
			current_frame.max = frames.length;
			frames.idx = 0;
			processNextFrame();
		}
		else {
			this.onprogress = null;
			error_handler(this.img_type + ": not animated...");
		}
	}
};

xhr.onerror = function() {
	this.onprogress = null;
	error_handler("Failed to load!");
};

xhr.open("GET", win.location.href, true);
xhr.send();